import fontforge
import shutil
import os
from tomlkit.toml_file import TOMLFile, TOMLDocument
from typing import Iterable, Optional

fontbakery_path = None
fontspector_path = None
fontbakery_config = TOMLDocument()
fontspector_config = TOMLDocument()

plugin_config = TOMLDocument()
_plugin_dir = ''
CONFIGFILE = 'config.toml'
FONTBAKERY_CONFIGFILE = 'fontbakery.toml'
FONTSPECTOR_CONFIGFILE = 'fontspector.toml'

profiles = {
    'opentype': 'OpenType (standards compliance)',
    'universal': 'Universal (community best practice)',
    'googlefonts': 'Google Fonts',
    'iso15008': 'ISO 15008 (in-car accessibility)',
    'fontwerk': 'Fontwerk',
    'adobefonts': 'Adobe Fonts',
    'fontbureau': 'Font Bureau',
    'microsoft': 'Microsoft',
    'notofonts': 'Noto fonts',
    'typenetwork': 'Type Network',
}
"""List of known Fontbakery/Fontspector check profile"""


def checkFontTools():
    global fontbakery_path, fontspector_path
    fontbakery_path = shutil.which('fontbakery')
    fontspector_path = shutil.which('fontspector')


def _validateConfItem(key: str, defaultVal, *, choice: Optional[Iterable] = None) -> bool:
    loaded = True
    if (key not in plugin_config) or (not isinstance(plugin_config[key], type(defaultVal))):
        plugin_config[key] = defaultVal  # load default
        loaded = False
    if choice:
        if not any(plugin_config[key] == x for x in choice):
            fontforge.logWarning("Invalid " + key + " '" + str(plugin_config[key]) + "' ignored")
            plugin_config[key] = defaultVal
    return loaded


def _validateConf():
    global profiles
    _validateConfItem('backend', 'auto', choice=['auto', 'fontbakery', 'fontspector'])
    _validateConfItem('check_as', 'ttf', choice=['ttf', 'ufo'])
    if _validateConfItem('profiles', profiles):
        profiles |= plugin_config['profiles']
    if _validateConfItem('profile', 'universal'):
        if plugin_config['profile'] not in profiles:
            profiles[plugin_config['profile']] = plugin_config['profile']
    _validateConfItem('explicit_checks', [])
    _validateConfItem('exclude_checks', [])


def fontBakeryConfigFile() -> str:
    return _plugin_dir + '/' + FONTBAKERY_CONFIGFILE


def fontSpectorConfigFile() -> str:
    return _plugin_dir + '/' + FONTSPECTOR_CONFIGFILE


def loadConf(confdir: str):
    global _plugin_dir, plugin_config, fontbakery_config, fontspector_config
    _plugin_dir = confdir
    try:
        plugin_config = TOMLFile(_plugin_dir + '/' + CONFIGFILE).read()
    except FileNotFoundError:
        pass

    _validateConf()

    try:
        fontbakery_config = TOMLFile(fontBakeryConfigFile()).read()
    except FileNotFoundError:
        pass
    try:
        fontspector_config = TOMLFile(fontSpectorConfigFile()).read()
    except FileNotFoundError:
        pass


def saveConf():
    os.makedirs(_plugin_dir, exist_ok=True)
    TOMLFile(_plugin_dir + '/' + CONFIGFILE).write(plugin_config)
    TOMLFile(_plugin_dir + '/' + FONTBAKERY_CONFIGFILE).write(fontbakery_config)
    TOMLFile(_plugin_dir + '/' + FONTSPECTOR_CONFIGFILE).write(fontspector_config)


def _writeBackendConf():
    fontspector_config['explicit_checks'] \
        = fontbakery_config['explicit_checks'] \
        = plugin_config['explicit_checks']
    fontspector_config['exclude_checks'] \
        = fontbakery_config['exclude_checks'] \
        = plugin_config['exclude_checks']
    for conf in (fontspector_config, fontbakery_config):
        for i in [x[0] for x in conf.items() if not x[1]]:
            conf.remove(i)


def configInterface():
    ans = fontforge.askMulti(
        'Configuration',
        [
            {
                'type': 'choice',
                'question': 'Backend',
                'tag': 'backend',
                'checks': True,
                'answers': [
                    {'name': p.capitalize(), 'tag': p, 'default': plugin_config['backend'] == p}
                    for p in ['auto', 'fontbakery', 'fontspector']
                ],
            },
            {
                'type': 'choice',
                'question': 'Check as',
                'tag': 'check_as',
                'checks': True,
                'answers': [
                    {'name': p, 'tag': p.lower(), 'default': plugin_config['check_as'] == p.lower()}
                    for p in ['TTF', 'UFO']
                ],
            },
            {
                'type': 'choice',
                'question': 'Profile',
                'tag': 'profile',
                'answers': [
                    {'name': p[1], 'tag': p[0], 'default': plugin_config['profile'] == p[0]}
                    for p in profiles.items()
                ],
            },
            {
                'type': 'string',
                'question': 'Explicit checks\n(comma-separated)',
                'tag': 'explicit_checks',
                'default': ','.join(plugin_config['explicit_checks']),
            },
            {
                'type': 'string',
                'question': 'Excluded checks\n(comma-separated)',
                'tag': 'exclude_checks',
                'default': ','.join(plugin_config['exclude_checks']),
            },
        ]
    )
    if ans:
        plugin_config['backend'] = ans['backend']
        plugin_config['check_as'] = ans['check_as']
        plugin_config['profile'] = ans['profile']
        plugin_config['explicit_checks'] = [a for a in (ans['explicit_checks'] or '').split(',') if a]
        plugin_config['exclude_checks'] = [a for a in (ans['exclude_checks'] or '').split(',') if a]
        _writeBackendConf()
        saveConf()



from . import config
import fontforge
import tempfile
from typing import Optional
from subprocess import run
import json
import webbrowser
from pathlib import Path

RESULT_JSON = 'lastresult.json'
RESULT_HTML = 'lastresult.html'


def _jsonFile() -> str:
    return config._plugin_dir + '/' + RESULT_JSON


def _htmlFile() -> str:
    return config._plugin_dir + '/' + RESULT_HTML


def _executable() -> Optional[str]:
    if config.plugin_config['backend'] == 'auto':
        return config.fontspector_path or config.fontbakery_path
    elif config.plugin_config['backend'] == 'fontbakery':
        return config.fontbakery_path
    elif config.plugin_config['backend'] == 'fontspector':
        return config.fontspector_path
    else:  # invalid!
        return None


def enabled(u, font) -> bool:
    return bool(_executable())


def _cmdline(filename: str) -> list[str]:
    if _executable():
        isFontSpector = (_executable() == config.fontspector_path)
        cmdline = [_executable()]
        if isFontSpector:
            cmdline.append('-p')
            cmdline.append(config.plugin_config['profile'])
        else:
            cmdline.append('check-' + config.plugin_config['profile'])
        cmdline.append('-q')
        cmdline.append('--full-lists')
        cmdline.append('-l')
        cmdline.append('info' if isFontSpector else 'INFO')
        cmdline.append('--configuration')
        cmdline.append(config.fontSpectorConfigFile() if isFontSpector else config.fontBakeryConfigFile())
        cmdline.append('--json')
        cmdline.append(_jsonFile())
        cmdline.append('--html')
        cmdline.append(_htmlFile())
        cmdline.append(filename)
        return cmdline
    else:
        raise RuntimeError('neither Fontbakery nor Fontspector available')


def _outroTitle(summary: dict) -> str:
    if 'ERROR' in summary and summary['ERROR'] > 0:
        return 'Error'
    elif 'FATAL' in summary and summary['FATAL'] > 0:
        return 'Check failed'
    elif 'FAIL' in summary and summary['FAIL'] > 0:
        return 'Check failed'
    else:
        return 'Check passed'


def _outroMessage(summary: dict) -> str:
    if 'ERROR' in summary and summary['ERROR'] > 0:
        return 'There ' + ('are errors' if summary['ERROR'] > 1 else 'is an error') + 'during check.'
    elif 'FATAL' in summary and summary['FATAL'] > 0:
        return 'Check failed with ' + ('severe issues' if summary['FATAL'] > 1 else 'a severe issue') + '.'
    elif 'FAIL' in summary and summary['FAIL'] > 0:
        return 'Check failed with ' + ('some issues' if summary['FAIL'] > 1 else 'an issue') + '.'
    elif 'WARN' in summary and summary['WARN'] > 0:
        return 'Check passed, but with ' + ('warnings' if summary['WARN'] > 1 else 'a warning') + '.'
    else:
        return 'Check passed.'


def _outroResultText(summary: dict) -> str:
    def toString(key: str, label: str) -> str:
        if key in summary and summary[key] > 0:
            return label + ' ' + str(summary[key])
        else:
            return label + ' 0'

    return ', '.join(
        toString(x[0], x[1]) for x in [
            ('ERROR', '💥'),
            ('FATAL', '☠'),
            ('FAIL', '🔥'),
            ('WARN', '⚠️'),
            ('INFO', 'ℹ️'),
            ('SKIP', '⏩'),
            ('PASS', '✅'),
        ]
    )


def _outro(filename: str):
    isFontSpector = (_executable() == config.fontspector_path)
    with open(_jsonFile(), 'r') as file:
        jsonDoc = json.load(file)
        summary = jsonDoc['summary'] if isFontSpector else jsonDoc['result']
        fontforge.logWarning(filename + ': ' + _outroResultText(summary))
        ans = fontforge.ask(
            _outroTitle(summary),
            _outroMessage(summary) + '\n'
            'Would you like to open details with the browser?',
            ['_Yes', '_No'])
        if ans == 0:
            webbrowser.open('file://' + _htmlFile(), 1)


def _run_check_direct(font: fontforge.font):
    run(_cmdline(font.path))
    _outro(Path(font.path).name)


def _run_check_tmpfile(font: fontforge.font):
    with tempfile.TemporaryDirectory() as tmpdir:
        changed = font.changed
        basename = (
            (font.default_base_filename or font.cidfontname or font.fontname) +
            '.' + config.plugin_config['check_as']
        )
        testfile = tmpdir + '/' + basename
        font.generate(testfile)
        font.changed = changed
        run(_cmdline(testfile))
        _outro(basename)


def run_check(u, font: fontforge.font):
    config.saveConf()
    if any(font.path.endswith(x) for x in ['.ttf', '.otf', '.ufo', '.ufo2', '.ufo3']):
        if font.changed:
            ans = fontforge.ask(
                'Font has been changed',
                'The font\n' + font.path + '\n'
                'has unsaved changes.\n'
                'How would you like to check?',
                ['Expor_t a temporary file', 'Check _existing font'],
            )
            if ans == 0:
                _run_check_tmpfile(font)
            else:
                _run_check_direct(font)
        else:
            _run_check_direct(font)
    else:
        _run_check_tmpfile(font)



import fontforge
from . import config, run_check


def fontforge_plugin_config(**_):
    config.configInterface()


def fontforge_plugin_init(preferences_path=None, **_):
    config.checkFontTools()
    config.loadConf(preferences_path)

    fontforge.registerMenuItem(
        callback=run_check.run_check,
        enable=run_check.enabled,
        context="Font",
        name="Check font"
    )
